from partitiontypes.PartitionTypes import PartitionTypes

class PartitionEntry():
    partitionTypes = PartitionTypes()
    def __init__(self, partitionBytes) -> None:
        # Boot indicator
        self.status = bytearray()                       #   1 Byte
        # Starting sector in CHS notation
        self.fSector = bytearray()                      #   3 Byte
        # Partition type
        self.partType = bytearray()                     #   1 Byte
        # Ending sector in CHS notation
        self.lSector = bytearray()                      #   3 Byte
        # offset to starting sector in LBA notation
        self.offset = bytearray()                       #   4 Byte
        # Number of sectors / size of partition
        self.noOfSectors = bytearray()                  #   4 Byte
        
        self.fSectorBin = ""
        self.lSectorBin = ""

        self.chs = {
            "start": {
                "head": 0,
                "sector": 0,
                "cylinder": 0,
                "headbin": "",
                "sectorbin": "",
                "cylinderbin": ""
            },
             "end": {
                "head": 0,
                "sector": 0,
                "cylinder": 0,
                "headbin": "",
                "sectorbin": "",
                "cylinderbin": ""
            }
        }

        self.__process(partitionBytes)


    def __process(self, pBytes):
        status = [pBytes[i:i + 1] for i in range(0, 1, 1)][0]
        self.status.extend(status)

        fSector = [pBytes[i:i + 3] for i in range(1, 4, 3)][0]
        self.fSector.extend(fSector)

        partType = [pBytes[i:i + 1] for i in range(4, 5, 1)][0]
        self.partType.extend(partType)

        lSector = [pBytes[i:i + 3] for i in range(5, 8, 3)][0]
        self.lSector.extend(lSector)

        offset = [pBytes[i:i + 4] for i in range(8, 12, 4)][0]
        self.offset.extend(offset)
        
        noOfSectors = [pBytes[i:i + 4] for i in range(12, 16, 4)][0]
        self.noOfSectors.extend(noOfSectors)

        h = self.fSector.hex(" ").replace(" ", "")
        self.fSectorBin = self.__hexToBin(h).rjust(24, "0")
        self.__calculateCHS("start", self.fSectorBin)

        h = self.lSector.hex(" ").replace(" ", "")
        self.lSectorBin = self.__hexToBin(h).rjust(24, "0")
        self.__calculateCHS("end", self.lSectorBin)

    def __toLittleEndian(self, value):
        # example for value = "20 21 00"
        s = value.split(" ")
        h_ = "".join(s[i] for i in range(len(s)-1, -1, -1))
        return h_

    def __hexToBin(self, hex):
        return bin(int(hex, base=16)).replace("0b","")

    def __hexToDec(self, hex_):
        return int(hex_, 16)

    def __binToDec(self, bin):
        # example for bin = "100001"
        return int(bin, 2)

    def __calculateCHS(self, type_, binary):
        if(len(binary) < 24):
            binary = binary.rjust(24, "0")

        self.chs[type_]["headbin"] = "".join(binary[i:i+8] for i in range(0, 8, 8))

        
        firstTwoCylBits = "".join(binary[i:i+2] for i in range(8, 10, 2))

        self.chs[type_]["sectorbin"] = "".join(binary[i:i+6] for i in range(10, 16, 6))

        lastSixCylBits = "".join(binary[i:i+8] for i in range(16, 24, 8))

        self.chs[type_]["cylinderbin"]  = firstTwoCylBits + lastSixCylBits

        self.chs[type_]["head"] = self.__binToDec(self.chs[type_]["headbin"])
        self.chs[type_]["sector"] = self.__binToDec(self.chs[type_]["sectorbin"])
        self.chs[type_]["cylinder"] = self.__binToDec(self.chs[type_]["cylinderbin"])


    def getOffsetLBA(self):
        h = self.offset.hex(" ")
        lEHex = self.__toLittleEndian(h)
        return self.__hexToDec(lEHex)

    def getNoOfSectors(self):
        h = self.noOfSectors.hex(" ")
        lEHex = self.__toLittleEndian(h)
        return self.__hexToDec(lEHex)

    def getNoOfBytes(self):
        return self.getNoOfSectors() * 512

    def getStatus(self):
        result = "Not bootable"
        if(self.status.hex() == "80"):
            result = "Bootable"
        return result

    def getPartitionTypes(self):
        t = self.partType.hex(" ").upper()
        return PartitionEntry.partitionTypes.getIds(t)


import os
import sys

import argparse

from hash_calc.HashCalc import HashCalc

from disk_detective.Controller import Controller

from disk_detective.mbr.MBRPrinter import MBRPrinter
from disk_detective.mbr.MasterBootRecord import MasterBootRecord

from disk_detective.structure.DiskStructurePrinter import DiskStructurePrinter

def main(args=None):
    """The main routine."""
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser()
    parser.add_argument("--path", "-p", type=str, required=True, help="Path to file")
    parser.add_argument("--mode", "-m", choices=["mbr", "structure"], required=True, help="The tool to use")
    args = parser.parse_args()

    c = Controller()
    hash = HashCalc(args.path)

    c.printHeader(args.path, hash)

    with open(args.path, "rb") as f:
        mbrBytes = f.read(512)
        mbr = MasterBootRecord(mbrBytes)

    if(args.mode == "mbr"):
            mbrPrinter = MBRPrinter()
            mbrPrinter.print(mbr)
    elif(args.mode == "structure"):
            structurePrinter = DiskStructurePrinter()
            structurePrinter.print(mbr)

    c.printExecutionTime()


if __name__ == "__main__":
    sys.exit(main())





class MBRPrinter():
    def __init__(self) -> None:
        pass

    def print(self, mbr):
        bl = mbr.bootloader.hex(" ").upper().split(" ")
        splitted = [bl[i:i + 16] for i in range(0, 440, 16)]
        print("")
        print("Master Boot Record Examiner")
        print("---------------------------")
        print("")
        print("--> Bootloader (0x0000 - 0x01B7 | 440 Bytes) ")
        self.__printBootloader(splitted)
        print("")
        print("--> Serial Number (0x01B8 - 0x01BB | 4 Bytes) ")
        print("    " + mbr.serialNumber.hex(" ").upper())
        print("")
        print("--> Null Bytes (0x01BC - 0x01BD | 2 Bytes) ")
        print("    " + mbr.nullBytes.hex(" ").upper())
        print("")
        print("--> Partition Entries (0x01BE - 0x01FD | 64 Bytes) ")
        for partitionEntry in mbr.partitionEntries:
            self.__printPartitionEntry(partitionEntry)

        print("--> Identifier (0x01FE - 0x01FF | 2 Bytes) ")
        print("    " + mbr.identifier.hex(" ").upper())

    def __printPartitionEntry(self, entry):
        print("")
        s = entry.status.hex(" ").upper()
        f = entry.fSector.hex(" ").upper()
        p = entry.partType.hex(" ").upper()
        partitionTypes = entry.getPartitionTypes()
        l = entry.lSector.hex(" ").upper()
        o = entry.offset.hex(" ").upper()
        n = entry.noOfSectors.hex(" ").upper()

        print("---------->      Entry: " + s + " " + f + " " + p + " " + l + " " + o + " " + n)
        print("                 -----")
        print("                Status: " + s)
        print("        --------------")
        print("                        " + entry.getStatus())
        print("        --------------")
        print("    First Sector (CHS): " + f)
        print("        --------------")
        print(f"            ---------->                {f} to bin: " + entry.fSectorBin[0:8] + " " + entry.fSectorBin[8:16] + " " + entry.fSectorBin[16:24])
        print("            ---------->         Head   " + entry.chs["start"]["headbin"] + " to dec: " + str(entry.chs["start"]["head"]))
        print("            ---------->       Sector     " + entry.chs["start"]["sectorbin"] + " to dec: " + str(entry.chs["start"]["sector"]))
        print("            ---------->     Cylinder " + entry.chs["start"]["cylinderbin"] + " to dec: " + str(entry.chs["start"]["cylinder"]))
        print("")
        print("                            CHS: " + str(entry.chs["start"]["cylinder"]) + "/" + str(entry.chs["start"]["head"]) + "/" + str(entry.chs["start"]["sector"]))
        print("        --------------")
        print("        Partition Type: " + p)
        print("        --------------")
        print("            Candidates:")
        for t in partitionTypes:
            print("                   ---> ID: " + t.id + "   Type: " + t.description)
        print("        --------------")
    
        print("     Last Sector (CHS): " + l)
        print("        --------------")
        print(f"            ---------->                {l} to bin: " + entry.lSectorBin[0:8] + " " + entry.lSectorBin[8:16] + " " + entry.lSectorBin[16:24])
        print("            ---------->         Head   " + entry.chs["end"]["headbin"] + " to dec: " + str(entry.chs["end"]["head"]))
        print("            ---------->       Sector     " + entry.chs["end"]["sectorbin"] + " to dec: " + str(entry.chs["end"]["sector"]))
        print("            ---------->     Cylinder " + entry.chs["end"]["cylinderbin"] + " to dec: " + str(entry.chs["end"]["cylinder"]))
        print("")
        print("                            CHS: " + str(entry.chs["end"]["cylinder"]) + "/" + str(entry.chs["end"]["head"]) + "/" + str(entry.chs["end"]["sector"]))
        print("        --------------")
        print("Offset to first sector: " + o)
        print("        --------------")
        print("                        " + str(entry.getOffsetLBA()))
        print("        --------------")
        print("        No. of sectors: " + n)
        print("        --------------")
        print("                        " + str(entry.getNoOfSectors()))
        print("                        " + str((entry.getNoOfBytes()/1024)/1024) + " MB")
        print("        --------------")
        print("")

    def __printBootloader(self, splitted):
        for line in splitted:
            print("    ", end="")
            for value in line:
                print(value + " ", end="")
            print("")


from disk_detective.mbr.PartitionEntry import PartitionEntry

class MasterBootRecord():
    def __init__(self, data) -> None:            
        self.bootloader = bytearray()            #  440 Bytes
        self.serialNumber = bytearray()          #    4 Bytes
        self.nullBytes = bytearray()             #    2 Bytes
        self.partitionEntries = []               # 4*16 Bytes
        self.identifier = bytearray()            #    2 Bytes
        self.__process(data)

    def __process(self, mbrBytes):
        bootloader = [mbrBytes[i:i + 440] for i in range(0, 440, 440)][0]
        self.bootloader.extend(bootloader)
        
        serialNumber = [mbrBytes[i:i + 4] for i in range(440, 444, 4)][0]
        self.serialNumber.extend(serialNumber)

        nullBytes = [mbrBytes[i:i + 2] for i in range(444, 446, 2)][0]
        self.nullBytes.extend(nullBytes)

        partitionEntries = [mbrBytes[i:i + 64] for i in range(446, 510, 64)][0]
        self.__processPartitionEntries(partitionEntries)
        
        identifier = [mbrBytes[i:i + 2] for i in range(510, 512, 2)][0]
        self.identifier.extend(identifier)

    def __processPartitionEntries(self, partitionBytes):
        splitted = [partitionBytes[i:i + 16] for i in range(0, 64, 16)]
        for entry in splitted:
            self.partitionEntries.append(PartitionEntry(entry))
            



class DiskStructurePrinter():
    def __init__(self) -> None:
        self.__list = []

    def print(self, mbr):
        print("")
        print("Disk Structure Overview")
        print("-----------------------")
        print("")
        counter = 0
        startSector = 0

        self.__printHeader()

        for i in range(0, len(mbr.partitionEntries)):
            entry = mbr.partitionEntries[i]
            lba = entry.getOffsetLBA()
            noOfSectors = entry.getNoOfSectors()
            partTypes = entry.getPartitionTypes()

            # Check if the first partition table entry starts at 0
            if(i == 0):
                if(lba != 0):
                    # If not insert a empty line with LBA is 0, end LBA is the LBA of the next entry - 1
                    self.__printLine("-/-", "-/-", 0, lba - 1, lba - 1)
                    self.__printLine(counter, partTypes, lba, lba + noOfSectors - 1, noOfSectors)

            
            # Check if a partition entry is empty
            elif(lba == 0 and noOfSectors == 0):
                    # In this case, start lba is the lba of the previous entry
                    # and the end lba is the lba of the next entry - 1
                    prevEndLba = mbr.partitionEntries[i-1].getOffsetLBA() + mbr.partitionEntries[i-1].getNoOfSectors()
                    nextStartingLba = mbr.partitionEntries[i+1].getOffsetLBA()
                    # It also needs to be checked if it is the last entry to avoid problems with i
                    self.__printLine("-/-", "Empty", prevEndLba + 1, nextStartingLba - 1, nextStartingLba - prevEndLba)
            else:
                lba = entry.getOffsetLBA()
                partTypes = entry.getPartitionTypes()
                self.__printLine(counter, partTypes, lba, lba + noOfSectors - 1, noOfSectors)

            counter += 1



    def __printHeader(self):
        formatStr = "| {:12} | {:30} | {:12} | {:12} | {:14} |".format("  Number #", " Possible FS Types",  " Start LBA", "   End LBA", " No. of Sectors")
        print("-"*len(formatStr))
        print(formatStr)
        print("-"*len(formatStr))

    
    def __printLine(self, counter, partitionTypes, startLba, endLba, noOfSectors):
        if(type(partitionTypes) == list and len(partitionTypes) > 1):
            # Multiline Print
            pType = partitionTypes[0].description[0:28] 
            formatStr1 = "| {:12} | {:30} | {:12} | {:12} | {:14} |".format(str(counter), pType, str(startLba), str(endLba), str(noOfSectors))
            print(formatStr1)
            for i in range(1, len(partitionTypes)):
                pType = partitionTypes[i].description[0:28] 
                formatStrN = "| {:12} | {:30} | {:12} | {:12} | {:14} |".format("", pType, "", "", "")
                print(formatStrN)
            print("-"*len(formatStr1))   
        else:
            # Single Line Print
            if(type(partitionTypes) == str):
                pType = partitionTypes
            else:
                pType = partitionTypes[0].description[0:28]

            formatStr = "| {:12} | {:30} | {:12} | {:12} | {:14} |".format(str(counter), pType, str(startLba), str(endLba), str(noOfSectors))
            print(formatStr)
            print("-"*len(formatStr))



from setuptools import setup, find_packages

with open("README.md", "r") as r:
    desc = r.read()

setup(
    name="disk_detective",            
    version="0.1.0",
    author="5f0",
    url="https://github.com/5f0ne/disk-detective",
    description="Provides tools to analyze disks on byte level",
    classifiers=[
        "Operating System :: OS Independent ",
        "Programming Language :: Python :: 3 ",
        "License :: OSI Approved :: MIT License "
    ],
    license="MIT",
    long_description=desc,
    long_description_content_type="text/markdown",
    package_dir={"": "src"},
    packages=find_packages(where='src'),
    install_requires=[
        "hash_calc",
        "partitiontypes"
    ],
    entry_points={
        "console_scripts": [
            "disk_detective = disk_detective.__main__:main"
        ]
    }
)